@@ -211,6 +211,7 @@ mkfunction(iswalpha)
 mkfunction(isascii)
 mkfunction(iswcntrl)
 mkfunction(iswdigit)
+mkfunction(iswprint)
 mkfunction(iswgraph)
 mkfunction(iswlower)
 mkfunction(iswupper)
@@ -231,6 +232,7 @@ static const char_type char_types[] =
   { ATOM_white,			     iswhite },
   { ATOM_cntrl,			     fiswcntrl },
   { ATOM_digit,			     fiswdigit },
+  { ATOM_print,			     fiswprint },
   { ATOM_graph,			     fiswgraph },
   { ATOM_lower,			     fiswlower },
   { ATOM_upper,			     fiswupper },
@@ -334,7 +336,7 @@ do_char_type(term_t chr, term_t class, control_t h, int how)
 
       if ( !(do_enum & ENUM_CLASS) )
       { if ( !PL_get_name_arity(class, &cn, &arity) ||
-	     !(cc = char_type_by_name(cn, arity)) )
+	     !(cc = char_type_by_name(cn, (int)arity)) )
 	  return PL_error("char_type", 2, NULL,
 			  ERR_TYPE, ATOM_char_type, class);
       }
