@@ -94,8 +94,9 @@ typedef enum
   ERR_UNDEFINED_PROC,		/* Definition def */
   ERR_DUPLICATE_KEY,		/* term_t */
   ERR_PERMISSION_SSU_DEF,	/* Definition def */
-  ERR_DETERMINISM		/* Definition def, atom_t expect,
+  ERR_DETERMINISM,		/* Definition def, atom_t expect,
 				   atom_t found, atom_t decl */
+  ERR_DET_GOAL			/* term_t goal, atom_t expect, atom_t found */
 } PL_error_code;
 
 #define MSG_ERRNO		((char *)(-1))
