@@ -83,6 +83,7 @@
 #define PL_KERNEL		1
 #include <inttypes.h>
 #include "pl-builtin.h"
+#include "pl-macros.h"
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 		      PROLOG SYSTEM OPTIONS
@@ -543,6 +544,13 @@ Arithmetic comparison
 #define NE 5
 #define EQ 6
 
+#define LT_C <
+#define GT_C >
+#define LE_C <=
+#define GE_C >=
+#define NE_C !=
+#define EQ_C ==
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Operator types.  NOTE: if you change OP_*, check operatorTypeToAtom()!
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
@@ -1616,7 +1624,7 @@ struct choice
 #endif
   union
   { struct clause_choice clause;	/* Next candidate clause */
-    Code	PC;			/* Next candidate program counter */
+    Code	pc;			/* Next candidate program counter */
     word        foreign;		/* foreign redo handle */
   } value;
 };
@@ -2405,10 +2413,12 @@ typedef struct debuginfo
   intptr_t	retryFrame;		/* Frame to retry (local stack offset) */
 } pl_debugstatus_t;
 
+#if O_DEBUG
 typedef struct internaldebuginfo
 { int depth;           /* how many nested DEBUG() calls we are in */
   const char *channel; /* string representation of the debug channel */
 } pl_internaldebugstatus_t;
+#endif
 
 #define FT_ATOM		0		/* atom feature */
 #define FT_BOOL		1		/* boolean feature (true, false) */
@@ -2452,7 +2462,8 @@ typedef enum plflag
   PLFLAG_TABLE_SHARED,			/* By default shared tabling */
   PLFLAG_RATIONAL,			/* Natural rational numbers */
   PLFLAG_DEBUG_ON_INTERRUPT,		/* Debug on Control-C */
-  PLFLAG_OPTIMISE_UNIFY			/* Move unifications in clauses */
+  PLFLAG_OPTIMISE_UNIFY,		/* Move unifications in clauses */
+  PLFLAG_SHIFT_CHECK			/* Check suspicious shifts */
 } plflag;
 
 typedef struct
