@@ -868,6 +868,7 @@ PRED_IMPL("$unload_file", 1, unload_file, 0)
 
 static void	fix_discontiguous(p_reload *r);
 static void	fix_metapredicate(p_reload *r);
+static void	fix_det(p_reload *r);
 
 #ifdef O_PLMT
 #define GEN_RELOAD (GEN_MAX-PL_thread_self())
@@ -891,7 +892,8 @@ startReconsultFile(SourceFile sf)
     r->pred_access_count = popNPredicateAccess(0);
     sf->reload = r;
 
-    LD->gen_reload = r->reload_gen;
+    LD->reload.generation = r->reload_gen;
+    LD->reload.nesting++;
 
     for(cell = sf->procedures; cell; cell = cell->next)
     { Procedure proc = cell->value;
@@ -1091,6 +1093,7 @@ assertProcedureSource(SourceFile sf, Procedure proc, Clause clause ARG_LD)
 
     if ( reload->number_of_clauses++ == 0 )
     { fix_discontiguous(reload);
+      fix_det(reload);
       fix_ssu(reload, clause);
     }
 
@@ -1270,6 +1273,17 @@ fix_discontiguous(p_reload *r)
 }
 
 
+static void
+fix_det(p_reload *r)
+{ Definition def = r->predicate;
+
+  if ( true(def, P_DET) && false(r, P_DET) )
+  { clear(def, P_DET);
+    freeCodesDefinition(def, TRUE);
+  }
+}
+
+
 static void
 fix_ssu(p_reload *r, Clause clause)
 { Definition def = r->predicate;
@@ -1581,7 +1595,8 @@ endReconsult(SourceFile sf)
     sf->reload = NULL;
     freeHeap(reload, sizeof(*reload));
 
-    LD->gen_reload = GEN_INVALID;
+    if ( --LD->reload.nesting == 0 )
+      LD->reload.generation = GEN_INVALID;
 
     pl_garbage_collect_clauses();
     if ( sendDelayedEvents(TRUE) < 0 )
