@@ -193,9 +193,8 @@ static void
 DbgPrintInstruction(LocalFrame FR, Code PC)
 { static LocalFrame ofr = NULL;		/* not thread-safe */
 
-  if ( DEBUGGING(MSG_VMI) )
+  IF_DEBUGGING(MSG_VMI)
   { GET_LD
-    ENTER_DEBUG(MSG_VMI)
 
     if ( ofr != FR )
     { Sfprintf(Serror, "#%ld at [%ld] predicate %s\n",
@@ -220,7 +219,6 @@ DbgPrintInstruction(LocalFrame FR, Code PC)
 
       Sdprintf("\t%4ld %s\n", (long)(PC-relto), codeTable[decode(*PC)].name);
     }
-    EXIT_DEBUG(MSG_VMI)
   }
 }
 
@@ -364,25 +362,6 @@ Brief description of the local stack-layout.  This stack contains:
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 
-		 /*******************************
-		 *	 LOCAL ALLOCATION	*
-		 *******************************/
-
-/* Note that lTop can be >= lMax when calling ENSURE_LOCAL_SPACE() */
-
-#define ENSURE_LOCAL_SPACE(bytes, ifnot) \
-	if ( unlikely(addPointer(lTop, (bytes)) > (void*)lMax) ) \
-        { int rc; \
-	  SAVE_REGISTERS(qid); \
-	  rc = growLocalSpace__LD(bytes, ALLOW_SHIFT PASS_LD); \
-	  LOAD_REGISTERS(qid); \
-	  if ( rc != TRUE ) \
-	  { rc = raiseStackOverflow(rc); \
-	    ifnot; \
-	  } \
-	}
-
-
 
 		 /*******************************
 		 *	    FOREIGN FRAME	*
@@ -2243,20 +2222,27 @@ copyFrameArguments(LocalFrame from, LocalFrame to, size_t argc ARG_LD)
 #ifdef O_DEBUG_BACKTRACK
 int backtrack_from_line;
 choice_type last_choice;
-#define GO(label) do { backtrack_from_line = __LINE__; goto label; } while(0)
+#define GO(helper) do { backtrack_from_line = __LINE__; VMH_GOTO(helper); } while(0)
 #else
-#define GO(label) goto label
+#define GO(helper) VMH_GOTO(helper)
 #endif
 
-#define FRAME_FAILED		GO(frame_failed)
-#define CLAUSE_FAILED		GO(clause_failed)
-#define BODY_FAILED		GO(body_failed)
-#ifdef O_DEBUG
-#define THROW_EXCEPTION		do { throwed_from_line = __LINE__; \
-				     goto b_throw; } while(0)
-#else
-#define THROW_EXCEPTION		goto b_throw
+#define FRAME_FAILED		GO(deep_backtrack)
+#define CLAUSE_FAILED		GO(shallow_backtrack)
+#define BODY_FAILED		GO(shallow_backtrack)
+#ifdef O_DEBUGGER
+#define TRACE_RETRY		VMH_GOTO(retry)
 #endif
+#ifdef O_ATTVAR
+#define ATTVAR_WAKEUP		VMH_GOTO(wakeup)
+#endif
+#ifdef O_DEBUG
+#define THROW_EXCEPTION		do { THROWED_FROM_LINE = __LINE__; \
+				     _THROW_EXCEPTION; } while(0)
+#else
+#define THROW_EXCEPTION		do { _THROW_EXCEPTION; } while(0)
+#endif
+#define _THROW_EXCEPTION	VMH_GOTO(b_throw)
 
 #ifdef O_PROFILE
 #define Profile(g) if ( unlikely(LD->profile.active) ) g
@@ -2864,50 +2850,253 @@ typedef enum
   uwrite				/* Unification in write mode */
 } unify_mode;
 
-#define IF_WRITE_MODE_GOTO(label) \
-	if ( umode == uwrite ) VMI_GOTO(label)
+/* HACK: uncomment the following line to enable function mode */
+/* #define O_VMI_FUNCTIONS 1 */
+/* #define VMI_REGISTER_VARIABLES 1 */
+/* #define VMI_USE_REGISTER_VARIABLES 1 */
+/* registers here MUST be in the list of callee-saved registers! */
+#define LD_REGISTER "rbx"
+#define REGFILE_REGISTER "r12"
 
-#define TRUST_CLAUSE(cref) \
-	umode = uread; \
-	CL    = cref; \
-	lTop  = (LocalFrame)(ARGP + cref->value.clause->variables); \
-	ENSURE_LOCAL_SPACE(LOCAL_MARGIN, THROW_EXCEPTION); \
-	if ( debugstatus.debugging ) \
-	  newChoice(CHP_DEBUG, FR PASS_LD); \
-	PC    = cref->value.clause->codes; \
-	NEXT_INSTRUCTION;
-#define TRY_CLAUSE(cref, cond, altpc) \
-	umode = uread; \
-	CL    = cref; \
-	lTop  = (LocalFrame)(ARGP + cref->value.clause->variables); \
-	ENSURE_LOCAL_SPACE(LOCAL_MARGIN, THROW_EXCEPTION); \
-	if ( cond ) \
-	{ Choice ch = newChoice(CHP_JUMP, FR PASS_LD); \
-	  ch->value.PC = altpc; \
-	} else if ( debugstatus.debugging ) \
-	{ newChoice(CHP_DEBUG, FR PASS_LD); \
-	} \
-	PC    = cref->value.clause->codes; \
-	NEXT_INSTRUCTION;
+#include "pentium.h"
 
+#if O_VMI_FUNCTIONS
+struct register_file;
+#define VMI_RETTYPE Code
+# if VMI_USE_REGISTER_VARIABLES
+#  define VMI_ARG_DECL Code PC
+#  define VMI_ARG_PASS PC
+# else
+#  define VMI_ARG_DECL Code PC, struct register_file *registers ARG_LD
+#  define VMI_ARG_PASS PC, registers PASS_LD
+# endif
+typedef VMI_RETTYPE (*vmi_instr)(VMI_ARG_DECL);
+# if VMI_REGISTER_VARIABLES
+register struct register_file *__reg_registers asm(REGFILE_REGISTER);
+register PL_local_data_t *__reg_ld asm(LD_REGISTER);
+# endif
+#endif
+
+/* All the registers used in PL_next_solution et al; there will always be a
+ * variable or macro called REGISTERS with the current active registers */
+typedef struct register_file
+{ unify_mode umode;			/* Unification mode */
+# define     UMODE	(REGISTERS.umode)
+  int        slow_unify;		/* B_UNIFY_FIRSTVAR */
+# define     SLOW_UNIFY	(REGISTERS.slow_unify)
+  qid_t      qid;			/* External query ID (argument to PL_next_solution()) */
+# define     QID	(REGISTERS.qid)
+  QueryFrame qf;			/* Query frame */
+# define     QF		(REGISTERS.qf)
+  LocalFrame fr;			/* current frame */
+# define     FR		(REGISTERS.fr)
+# define     CL		(FR->clause)	/* clause of current frame */
+  LocalFrame nfr;			/* Next frame */
+# define     NFR	(REGISTERS.nfr)
+  Word       argp;			/* current argument pointer */
+# define     ARGP	(REGISTERS.argp)
+  Definition def;			/* definition of current procedure */
+# define     DEF	(REGISTERS.def)
+  fid_t      ffr_id;			/* foreign function id */
+# define     FFR_ID	(REGISTERS.ffr_id)
+  ar_context pl_ar_ctx;
+# define     __PL_ar_ctx (REGISTERS.pl_ar_ctx)
+  struct foreign_context fndet_context;	/* foreign function non-deterministic context */
+# define     FNDET_CONTEXT (REGISTERS.fndet_context)
+#ifdef O_DEBUG
+  int	     throwed_from_line;		/* Debugging: line we came from */
+# define     THROWED_FROM_LINE	(REGISTERS.throwed_from_line)
+#endif
+#if O_VMI_FUNCTIONS
+  int        solution_ret;		/* return value for PL_next_solution, when exit_vm_buf is used */
+# define     SOLUTION_RET		(REGISTERS.solution_ret)
+  jmp_buf    exit_vm_buf;		/* jump target for exiting PL_next_solution */
+# define     EXIT_VM_BUF		(REGISTERS.exit_vm_buf)
+#endif
+} register_file;
+
+
+/* Imperative code that gets executed just after entry to an instruction,
+ * and just before exit from an instruction (i.e. before goto/return/etc).
+ * For profiling/tracing purposes only, and not applied to VMH's.
+ */
+#define VMI_ENTER(n)		count(n, PC); START_PROF(n, #n);
+#define VMI_EXIT		END_PROF();
+
+/* Components of VMI/VMH macro expansion. The underscore-prefix macros
+ * get defined per-implementation.
+ */
+
+/* If the compiler complains about undeclared __is_vmi or __is_vmh, it means
+ * a mismatch of VMI..END_VMH or VMH..END_VMI.
+ */
+#define assert_exists(var, message) (void)(var)
+#define VMI(Name,f,na,a)	_VMI_DECLARATION(Name,f,na,a) \
+				{ int __is_vmi = 1; \
+				  { _VMI_PROLOGUE(Name,f,na,a); VMI_ENTER(Name)
+#define END_VMI			    _VMI_EPILOGUE \
+				  } \
+				  assert_exists(__is_vmi, "END_VMI used without VMI!"); \
+				}
+#define VMH(Name,na,at,an)	_VMH_DECLARATION(Name,na,at,an) \
+				{ int __is_vmh = 1; \
+				  { _VMH_PROLOGUE(Name,na,at,an)
+#define END_VMH			    _VMH_EPILOGUE \
+				  } \
+				  assert_exists(__is_vmh, "END_VMH used without VMH!"); \
+				}
+#define NEXT_INSTRUCTION	do { VMI_EXIT; _NEXT_INSTRUCTION; } while(0)
+#define VMI_GOTO(n)		do { VMI_EXIT; _VMI_GOTO(n); } while(0)
+#define VMH_GOTO(...)		do { _VMH_GOTO(__VA_ARGS__); } while(0)
+#define SOLUTION_RETURN(val)	do { VMI_EXIT; _SOLUTION_RETURN(val); } while(0)
+#define VMI_GOTO_CODE(c)	do { VMI_EXIT; _VMI_GOTO_CODE(c); } while(0)
+#define SEPARATE_VMI		(void)0 /* only needed for !O_VMI_FUNCTIONS && VMCODE_IS_ADDRESS */
+
+/* By default, instruction and helper prologue/epilogue are empty */
+#define _VMI_PROLOGUE(Name,f,na,a)	;
+#define _VMI_EPILOGUE			;
+#define _VMH_PROLOGUE(Name,na,at,an)	;
+#define _VMH_EPILOGUE			;
+
+/* Same syntax as VMH_GOTO, but handles profiling as if it were an instruction */
+#define VMH_GOTO_AS_VMI(n,args...) do { VMI_EXIT; \
+					VMI_ENTER(n); \
+					VMH_GOTO(n, args); \
+				      } while(0)
+
+
+/* Helper macros for rendering VMH arguments */
+#define HEAD(h, ...) h
+#define TAIL(_, t...) (t)
+#define VMH_ARGS0(n,at,an,f,asep...)
+#define VMH_ARGS1(n,at,an,f,asep...) f(n, HEAD at, HEAD an)
+#define VMH_ARGS2(n,at,an,f,asep...) f(n, HEAD at, HEAD an) asep VMH_ARGS1(n, TAIL at, TAIL an, f, asep)
+#define VMH_ARGS3(n,at,an,f,asep...) f(n, HEAD at, HEAD an) asep VMH_ARGS2(n, TAIL at, TAIL an, f, asep)
+#define VMH_ARGS4(n,at,an,f,asep...) f(n, HEAD at, HEAD an) asep VMH_ARGS3(n, TAIL at, TAIL an, f, asep)
+#define COMMA_TYPE_ARG(n,at,an) , at an
+#define TYPE_ARG_SEMI(n,at,an)	at an ;
+#define VMH_ARGS(n) A_PASTE(VMH_ARGS, VMH_ARGCOUNT(n))(n, (VMH_ARGTYPES(n)), (VMH_ARGNAMES(n)), TYPE_ARG_SEMI)
+
+/* Define struct types for all the helper argument lists */
+#define VMH_ARGSTRUCT(Name)		struct helper_args_ ## Name
+
+FOREACH_VMH(T_EMPTY,
+  ,VMH_ARGSTRUCT, { ,VMH_ARGS, };
+)
+
+#define ASSIGN_ARG(n,at,an)		at an = HELPER_ARGS(n).an;
+#undef _VMH_PROLOGUE
+#define _VMH_PROLOGUE(Name,na,at,an)	VMH_ARGS ## na(Name, at, an, ASSIGN_ARG)
+#if O_VMI_FUNCTIONS
+
+#define HELPER_ARGS(n)			__args
+#define _VMI_DECLARATION(Name,na,at,an)	static VMI_RETTYPE instr_ ## Name(VMI_ARG_DECL)
+#define _VMH_DECLARATION(Name,na,at,an)	static VMI_RETTYPE helper_ ## Name(VMI_ARG_DECL, VMH_ARGSTRUCT(Name) __args)
+#define _NEXT_INSTRUCTION		return PC
+#define _SOLUTION_RETURN(val)		SOLUTION_RET = (val); longjmp(EXIT_VM_BUF, 1)
+#define _VMI_GOTO(n)			PC--; return instr_ ## n(VMI_ARG_PASS)
+#define _VMH_GOTO(n,...)		VMH_ARGSTRUCT(n) __args = {__VA_ARGS__}; (void)__args; \
+					return helper_ ## n(VMI_ARG_PASS, __args)
+#undef _VMI_PROLOGUE
+#define _VMI_PROLOGUE(Ident,f,na,a)	PC++;
+#if VMCODE_IS_ADDRESS
+#define VMI_ADDR(c)		((vmi_instr)(c))
+#else
+#define VMI_ADDR(c)		jmp_table[c]
+#endif
+#define _VMI_GOTO_CODE(c)	PC--; return VMI_ADDR(c)(VMI_ARG_PASS)
+
+/* Declare prototypes for all VMI/VMH functions */
+FOREACH_VMIDECL_CALL(T_SEMICOLON, _VMI_DECLARATION);
+FOREACH_VMHDECL_CALL(T_SEMICOLON, _VMH_DECLARATION);
+
+/* Define the jump table with all the function addresses */
+static vmi_instr jmp_table[] =
+{ FOREACH_VMI(T_COMMA,
+    &,VMI_IDENT,
+  ),
+  NULL
+};
+
+/* Define implementations */
+#if VMI_USE_REGISTER_VARIABLES
+# define REGISTERS (*__reg_registers)
+# undef LD
+# define LD (__reg_ld)
+# define _GET_LD /* empty */
+#else
+# define REGISTERS (*registers)
+#endif
+#include "pl-vmi.c"
+#undef REGISTERS
+
+/* Redefine NEXT_INSTRUCTION and VMH_GOTO for PL_next_solution() */
+#undef NEXT_INSTRUCTION
+#undef VMH_GOTO
+#define NEXT_INSTRUCTION (void)0
+#define VMH_GOTO(n) PC = helper_##n(VMI_ARG_PASS, (VMH_ARGSTRUCT(n)){})
+
+#if VMI_USE_REGISTER_VARIABLES
+# undef LD
+# define LD LOCAL_LD
+#endif
+
+#else /* O_VMI_FUNCTIONS */
+
+#define HELPER_ARGS(n)			helper_args.n
+#define _VMH_DECLARATION(Name,na,at,an)	helper_ ## Name:
+#define _VMH_GOTO(n,args...)		VMH_ARGSTRUCT(n) __args = {args}; \
+					HELPER_ARGS(n) = __args; \
+					goto helper_ ## n;
+#define _SOLUTION_RETURN		return
+
+#if VMCODE_IS_ADDRESS
+
+#define _VMI_DECLARATION(Name,f,na,a)	Name ## _LBL:
+#define _NEXT_INSTRUCTION		DbgPrintInstruction(FR, PC); _VMI_GOTO_CODE(*PC++)
+#define _VMI_GOTO(n)			goto n ## _LBL
+#define _VMI_GOTO_CODE(c)		goto *(void *)(c)
+#undef SEPARATE_VMI
+/* This macro must ensure that two identical VMI instructions do not get
+ * merged onto the same address by the compiler, causing decompilation
+ * which translates the addresses back into the VMI number to fail.
+ * initWamTable() verfies this does not happen.
+ */
+#define SEPARATE_VMI { static volatile int nop = 0; (void)nop; }
+
+#else /* VMCODE_IS_ADDRESS */
+
+#if __GNUC__
+#define UNUSED_LABEL __attribute__ ((unused))
+#else
+#define UNUSED_LABEL
+#endif
+
+#define _VMI_DECLARATION(Name,f,na,a)	case Name: case_ ## Name: UNUSED_LABEL
+#define _NEXT_INSTRUCTION		goto next_instruction
+#define _VMI_GOTO(n)			goto case_ ## n
+#define _VMI_GOTO_CODE(c)		thiscode = (c); goto resumebreak;
+
+#endif /* VMCODE_IS_ADDRESS */
+#endif /* O_VMI_FUNCTIONS */
 
 int
 PL_next_solution(qid_t qid)
 { GET_LD
-  AR_CTX
-  QueryFrame QF;			/* Query frame */
-  LocalFrame FR;			/* current frame */
-  LocalFrame NFR;			/* Next frame */
-  Word	     ARGP;			/* current argument pointer */
-  Code	     PC = NULL;			/* program counter */
-  Definition DEF = NULL;		/* definition of current procedure */
-  unify_mode umode = uread;		/* Unification mode */
-  int slow_unify = FALSE;		/* B_UNIFY_FIRSTVAR */
-  exception_frame throw_env;		/* PL_thow() environment */
-#ifdef O_DEBUG
-  int	     throwed_from_line=0;	/* Debugging: line we came from */
-#endif
-#define	     CL (FR->clause)		/* clause of current frame */
+  register_file REGISTERS = {.qid = qid};	/* Active registers */
+  Code PC;					/* program counter */
+  exception_frame THROW_ENV;			/* PL_thow() environment */
+
+#if O_VMI_FUNCTIONS
+  register_file *registers = &REGISTERS;
+
+#else /* O_VMI_FUNCTIONS */
+  /* define local union with all "helper arguments" (formerly SHAREDVARS) */
+  union
+  { FOREACH_VMH(T_EMPTY,
+      ,VMH_ARGSTRUCT, ,VMH_NAME,;
+    )
+  } helper_args;
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Get the labels of the various  virtual-machine instructions in an array.
@@ -2918,68 +3107,25 @@ will export jmp_table as the compiler  needs   to  know  this table. See
 pl-comp.c
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#include "pentium.h"
 
 #if VMCODE_IS_ADDRESS
-#include <pl-jumptable.ic>
-
-#define VMI(Name,f,na,a)	Name ## _LBL: \
-				  count(Name, PC); \
-				  START_PROF(Name, #Name);
-#define VMI_GOTO(n)		do { END_PROF(); \
-				       goto n ## _LBL; \
-				   } while(0)
-#define NEXT_INSTRUCTION	do { END_PROF(); \
-				     DbgPrintInstruction(FR, PC); \
-				     goto *(void *)((intptr_t)(*PC++)); \
-				   } while(0)
-
-/* This macro must ensure that two identical VMI instructions do not get
- * merged onto the same address by the compiler, causing decompilation
- * which translates the addresses back into the VMI number to fail.
- * initWamTable() verfies this does not happen.
- */
-#define SEPARATE_VMI { static volatile int nop = 0; (void)nop; }
-
+#undef VMI_IDENT
+#define VMI_IDENT(n) n ## _LBL
+  static void *jmp_table[] =
+  { FOREACH_VMI(T_COMMA,
+      &&,VMI_IDENT,
+    ),
+    NULL
+  };
 #else /* VMCODE_IS_ADDRESS */
-
 code thiscode;
-
-#if __GNUC__
-#define UNUSED_LABEL __attribute__ ((unused))
-#else
-#define UNUSED_LABEL
-#endif
-
-#define VMI(Name,f,na,a)	case Name: \
-				  case_ ## Name: UNUSED_LABEL \
-				  count(Name, PC); \
-				  START_PROF(Name, #Name);
-#define VMI_GOTO(n)		{ END_PROF(); \
-				  goto case_ ## n; \
-				}
-#define NEXT_INSTRUCTION	{ DbgPrintInstruction(FR, PC); \
-				  END_PROF(); \
-                                  goto next_instruction; \
-				}
-#define SEPARATE_VMI		(void)0
-
 #endif /* VMCODE_IS_ADDRESS */
 
-#define FASTCOND_FAILED \
-	{ if ( !LD->fast_condition )   \
-	  { BODY_FAILED;	       \
-	  } else		       \
-	  { PC = LD->fast_condition;   \
-	    LD->fast_condition = NULL; \
-	    NEXT_INSTRUCTION;          \
-	  }                            \
-	}
-
+#endif /* O_VMI_FUNCTIONS */
 
 #if VMCODE_IS_ADDRESS
   if ( qid == QID_EXPORT_WAM_TABLE )
-  { interpreter_jmp_table = jmp_table;	/* make it globally known */
+  { interpreter_jmp_table = (void**) jmp_table;	/* make it globally known */
     succeed;
   }
 #endif /* VMCODE_IS_ADDRESS */
@@ -3006,6 +3152,21 @@ depart_continue() to do the normal thing or to the backtrack point.
   ARGP = argFrameP(FR, 0);
   DEBUG(9, Sdprintf("QF=%p, FR=%p\n", QF, FR));
 
+#if O_VMI_FUNCTIONS
+  if (setjmp(EXIT_VM_BUF) != 0)
+  { // LD->vmi_registers = old_registers;
+    assert(LD->exception.throw_environment == &THROW_ENV);
+    LD->exception.throw_environment = THROW_ENV.parent;
+    return SOLUTION_RET;
+  }
+
+# if VMI_REGISTER_VARIABLES
+  /* Now that we've executed our exit setjmp, we can set register vars with impunity */
+  __reg_registers = registers;
+  __reg_ld = LD;
+# endif
+#endif /*O_VMI_FUNCTIONS*/
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Check for exceptions raised by foreign code.  PL_throw() uses longjmp()
 to get back here.  Our task is to restore the environment and throw the
@@ -3017,8 +3178,8 @@ variables used in the B_THROW instruction.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
   DEBUG(9, Sdprintf("Setjmp env at %p\n", &LD->exception.throw_environment));
-  throw_env.parent = LD->exception.throw_environment;
-  if ( setjmp(throw_env.exception_jmp_env) != 0 )
+  THROW_ENV.parent = LD->exception.throw_environment;
+  if ( setjmp(THROW_ENV.exception_jmp_env) != 0 )
   { FliFrame ffr;
 #ifdef O_PLMT
     __PL_ld = GLOBAL_LD;		/* might be clobbered */
@@ -3040,8 +3201,8 @@ variables used in the B_THROW instruction.
 
     THROW_EXCEPTION;
   } else				/* installation */
-  { throw_env.magic = THROW_MAGIC;
-    LD->exception.throw_environment = &throw_env;
+  { THROW_ENV.magic = THROW_MAGIC;
+    LD->exception.throw_environment = &THROW_ENV;
   }
 
   DEF = FR->predicate;
@@ -3055,10 +3216,10 @@ variables used in the B_THROW instruction.
       if ( exception_term )
 	THROW_EXCEPTION;
       NEXT_INSTRUCTION;
-    }
-    BODY_FAILED;
+    } else
+      BODY_FAILED;
   } else
-    goto retry_continue;		/* first call */
+    VMH_GOTO(depart_or_retry_continue);		/* first call */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Main entry of the virtual machine cycle.  A branch to `next instruction'
@@ -3066,9 +3227,21 @@ will  cause  the  next  instruction  to  be  interpreted.   All  machine
 registers  should  hold  valid  data  and  the  machine stacks should be
 initialised properly.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-
+#if O_VMI_FUNCTIONS
+  for (;;)
+  { DbgPrintInstruction(FR, PC);
+#if VMI_REGISTER_VARIABLES && !VMI_USE_REGISTER_VARIABLES
+    DEBUG(0,
+      assert(__reg_registers == &REGISTERS);
+      assert(__reg_ld == LD);
+    );
+#endif
+    PC = VMI_ADDR(*PC)(VMI_ARG_PASS);
+  }
+#else /* O_VMI_FUNCTIONS */
 #if !VMCODE_IS_ADDRESS			/* no goto *ptr; use a switch */
 next_instruction:
+  DbgPrintInstruction(FR, PC);
   thiscode = *PC++;
 #ifdef O_DEBUGGER
 resumebreak:
@@ -3079,449 +3252,8 @@ resumebreak:
 #include "pl-vmi.c"
   }
 
-#ifdef O_ATTVAR
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Attributed variable handling
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-wakeup:
-  DEBUG(1, Sdprintf("Activating wakeup\n"));
-  NFR = lTop;
-  setNextFrameFlags(NFR, FR);
-  SAVE_REGISTERS(qid);
-  DEF = GD->procedures.dwakeup1->definition;
-  LOAD_REGISTERS(qid);
-  ARGP = argFrameP(NFR, 0);
-  ARGP[0] = *valTermRef(LD->attvar.head);
-  setVar(*valTermRef(LD->attvar.head));
-  setVar(*valTermRef(LD->attvar.tail));
+#endif /* O_VMI_FUNCTIONS */
 
-  goto normal_call;
-#endif
-
-
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-			TRACER RETRY ACTION
-
-By default, retries the  current  frame.  If   another  frame  is  to be
-retried, place the frame-reference, which  should   be  a  parent of the
-current frame, in debugstatus.retryFrame and jump to this label. This is
-implemented by returning retry(Frame) of the prolog_trace_interception/3
-hook.
-
-First, the system will leave any parent  frames. Next, it will undo back
-to the call-port and finally, restart the clause.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-
-#if O_DEBUGGER
-retry:					MARK(RETRY);
-{ LocalFrame rframe0, rframe;
-  mark m;
-  Choice ch;
-
-  if ( debugstatus.retryFrame )
-    rframe = (LocalFrame)valTermRef(debugstatus.retryFrame);
-  else
-    rframe = FR;
-  debugstatus.retryFrame = 0;
-  rframe0 = rframe;
-
-  m.trailtop = tTop;
-  m.globaltop = gTop;
-  for( ; rframe; rframe = rframe->parent )
-  { if ( (ch = findStartChoice(rframe, BFR)) )
-    { m = ch->mark;
-      goto do_retry;
-    }
-  }
-  Sdprintf("[Could not find retry-point]\n");
-  SAVE_REGISTERS(qid);
-  abortProlog();				/* What else? */
-  LOAD_REGISTERS(qid);
-  THROW_EXCEPTION;
-
-do_retry:
-  if ( rframe0 != rframe )
-  { DEBUG(MSG_TRACE,
-	  Sdprintf("[No retry-information for requested frame]\n"));
-  }
-
-  DEBUG(MSG_TRACE,
-	Sdprintf("[Retrying frame %d running %s]\n",
-		 (Word)rframe - (Word)lBase,
-		 predicateName(rframe->predicate)));
-
-  discardChoicesAfter(rframe, FINISH_CUT PASS_LD);
-  rframe->clause = NULL;
-  environment_frame = FR = rframe;
-  DEF = FR->predicate;
-  clear(FR, FR_SKIPPED);
-  Undo(m);
-  exception_term = 0;
-
-  goto retry_continue;
-}
-#endif /*O_DEBUGGER*/
-
-		 /*******************************
-		 *	   BACKTRACKING		*
-		 *******************************/
-
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-The rest of this giant procedure handles   backtracking. This used to be
-very complicated, but as of pl-3.3.6, choice-points are explicit objects
-and life is a lot easier. In the old days we distinquished between three
-cases to get here. We leave that   it for documentation purposes as well
-as to investigate optimization in the future.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-
-				MARK(BKTRK);
-clause_failed:				/* shallow backtracking */
-body_failed:
-END_PROF();
-START_PROF(P_SHALLOW_BACKTRACK, "P_SHALLOW_BACKTRACK");
-{ Choice ch = BFR;
-
-  if ( FR == ch->frame )
-  { Undo(ch->mark);
-    QF = QueryFromQid(qid);
-    aTop = QF->aSave;
-
-    if ( ch->type == CHP_JUMP )
-    { DiscardMark(ch->mark);
-      PC   = ch->value.PC;
-      BFR  = ch->parent;
-      lTop = (LocalFrame)ch;
-      ARGP = argFrameP(lTop, 0);
-
-      NEXT_INSTRUCTION;
-    } else if ( ch->type == CHP_CLAUSE )
-    { ARGP = argFrameP(FR, 0);
-      if ( !(CL = nextClause(&ch->value.clause, ARGP, FR, DEF)) )
-	FRAME_FAILED;		/* can happen if scan-ahead was too short */
-      PC = CL->value.clause->codes;
-      umode = uread;
-
-      if ( ch == (Choice)argFrameP(FR, CL->value.clause->variables) )
-      { DiscardMark(ch->mark);		/* is this needed? */
-	if ( ch->value.clause.cref )
-	{ Mark(ch->mark);
-	  lTop = (LocalFrame)(ch+1);
-	  NEXT_INSTRUCTION;
-	} else if ( unlikely(debugstatus.debugging) )
-	{ ch->type = CHP_DEBUG;
-	  Mark(ch->mark);
-	  lTop = (LocalFrame)(ch+1);
-	  NEXT_INSTRUCTION;
-	}
-
-	BFR = ch->parent;
-	lTop = (LocalFrame)ch;
-	NEXT_INSTRUCTION;
-      } else				/* Choice point needs to move */
-      { struct clause_choice chp;
-
-        DiscardMark(ch->mark);
-	BFR = ch->parent;
-	chp = ch->value.clause;
-	lTop = (LocalFrame)argFrameP(FR, CL->value.clause->variables);
-	ENSURE_LOCAL_SPACE(LOCAL_MARGIN, THROW_EXCEPTION);
-
-	if ( chp.cref )
-	{ ch = newChoice(CHP_CLAUSE, FR PASS_LD);
-	  ch->value.clause = chp;
-	} else if ( unlikely(debugstatus.debugging) )
-	{ ch = newChoice(CHP_DEBUG, FR PASS_LD);
-	}
-	NEXT_INSTRUCTION;
-      }
-    }
-  }
-}
-
-
-frame_failed:
-END_PROF();
-START_PROF(P_DEEP_BACKTRACK, "P_DEEP_BACKTRACK");
-{
-#ifdef O_DEBUGGER
-  term_t ch0_ref = BFR ? consTermRef(BFR) : 0;
-#endif
-  Choice ch;
-
-  DEBUG(MSG_BACKTRACK, Sdprintf("BACKTRACKING\n"));
-
-next_choice:
-  ch = BFR;
-					/* leave older frames */
-  for(; (void *)FR > (void *)ch; FR = FR->parent)
-  {
-#ifdef O_DEBUGGER
-    if ( debugstatus.debugging && isDebugFrame(FR) )
-    { Choice sch = ch0_ref ? findStartChoice(FR, (Choice)valTermRef(ch0_ref)) : NULL;
-
-      DEBUG(MSG_BACKTRACK,
-	    Sdprintf("FAIL on %s\n", predicateName(FR->predicate)));
-
-      if ( sch )
-      { int rc;
-	Choice ch0 = findChoiceBeforeFrame(FR, sch);
-
-	ch0_ref = ch0 ? consTermRef(ch0) : 0;
-	Undo(sch->mark);
-	environment_frame = FR;
-	FR->clause = NULL;
-	lTop = (LocalFrame)argFrameP(FR, FR->predicate->functor->arity);
-	SAVE_REGISTERS(qid);
-	rc = tracePort(FR, BFR, FAIL_PORT, NULL PASS_LD);
-	LOAD_REGISTERS(qid);
-	ch = BFR;			/* can be shifted */
-
-	switch( rc )
-	{ case ACTION_RETRY:
-	    environment_frame = FR;
-	    DEF = FR->predicate;
-	    clear(FR, FR_CATCHED|FR_SKIPPED);
-	    goto retry_continue;
-	    case ACTION_ABORT:
-	      THROW_EXCEPTION;
-	}
-      } else
-      { ch0_ref = 0;
-	DEBUG(2, Sdprintf("Cannot trace FAIL [%d] %s\n",
-			  levelFrame(FR), predicateName(FR->predicate)));
-      }
-    }
-#endif
-
-    leaveFrame(FR);
-    if ( true(FR, FR_WATCHED|FR_SSU_DET|FR_DET|FR_DETGUARD) )
-    { environment_frame = FR;
-      lTop = (LocalFrame)argFrameP(FR, FR->predicate->functor->arity);
-      FR->clause = NULL;
-      if ( true(FR, FR_SSU_DET|FR_DET|FR_DETGUARD) )
-      { SAVE_REGISTERS(qid);
-	ssu_or_det_failed(FR PASS_LD);
-	LOAD_REGISTERS(qid);
-	if ( exception_term )
-	  THROW_EXCEPTION;
-      }
-      SAVE_REGISTERS(qid);
-      frameFinished(FR, FINISH_FAIL PASS_LD);
-      LOAD_REGISTERS(qid);
-      ch = BFR;			/* can be shifted */
-      if ( exception_term )
-	THROW_EXCEPTION;
-    }
-  }
-
-  environment_frame = FR = ch->frame;
-  Undo(ch->mark);
-  QF = QueryFromQid(qid);
-  aTop = QF->aSave;
-  DEF  = FR->predicate;
-#ifdef O_DEBUG_BACKTRACK
-  last_choice = ch->type;
-#endif
-
-  if ( LD->alerted )
-  { if ( LD->alerted & ALERT_BUFFER )
-    { LD->alerted &= ~ALERT_BUFFER;
-      release_string_buffers_from_frame(FR PASS_LD);
-    }
-    if ( UNDO_SCHEDULED(LD) )
-    { int rc;
-
-      SAVE_REGISTERS(qid);
-      rc = run_undo_hooks(PASS_LD1);
-      LOAD_REGISTERS(qid);
-      if ( !rc )
-	THROW_EXCEPTION;
-    }
-  }
-
-  switch(ch->type)
-  { case CHP_JUMP:
-      DEBUG(MSG_BACKTRACK,
-	    Sdprintf("    REDO #%ld: Jump in %s\n",
-		     loffset(FR),
-		     predicateName(DEF)));
-      PC   = ch->value.PC;
-      DiscardMark(ch->mark);
-      BFR  = ch->parent;
-      lTop = (LocalFrame)ch;
-      ARGP = argFrameP(lTop, 0);
-      LD->statistics.inferences++;
-      if ( unlikely(LD->alerted) )
-      {
-#ifdef O_DEBUGGER
-	if ( debugstatus.debugging && !debugstatus.suspendTrace  )
-	{ LocalFrame fr = dbgRedoFrame(FR, CHP_JUMP PASS_LD);
-
-	  if ( fr )
-	  { int action;
-
-	    SAVE_REGISTERS(qid);
-	    action = tracePort(fr, BFR, REDO_PORT, ch->value.PC PASS_LD);
-	    LOAD_REGISTERS(qid);
-	    ch = BFR;			/* can be shifted */
-
-	    switch( action )
-	    { case ACTION_FAIL:
-		FRAME_FAILED;
-	      case ACTION_IGNORE:
-		VMI_GOTO(I_EXIT);
-	      case ACTION_RETRY:
-	        goto retry;
-	      case ACTION_ABORT:
-		THROW_EXCEPTION;
-	    }
-	  }
-	}
-#endif
-#ifdef O_INFERENCE_LIMIT
-        if ( LD->statistics.inferences >= LD->inference_limit.limit )
-	{ SAVE_REGISTERS(qid);
-	  raiseInferenceLimitException();
-	  LOAD_REGISTERS(qid);
-	  if ( exception_term )
-	    THROW_EXCEPTION;
-	}
-#endif
-        Profile(profRedo(ch->prof_node PASS_LD));
-      }
-      NEXT_INSTRUCTION;
-    case CHP_CLAUSE:			/* try next clause */
-    { Clause clause;
-      struct clause_choice chp;
-
-      DEBUG(MSG_BACKTRACK,
-	    Sdprintf("    REDO #%ld: Clause in %s\n",
-		     loffset(FR),
-		     predicateName(DEF)));
-      ARGP = argFrameP(FR, 0);
-      DiscardMark(ch->mark);
-      BFR = ch->parent;
-      if ( !(CL = nextClause(&ch->value.clause, ARGP, FR, DEF)) )
-	goto next_choice;	/* Can happen of look-ahead was too short */
-
-      chp    = ch->value.clause;
-      clause = CL->value.clause;
-      PC     = clause->codes;
-      lTop   = (LocalFrame)argFrameP(FR, clause->variables);
-      umode  = uread;
-
-      DEBUG(CHK_SECURE, assert(LD->mark_bar >= gBase && LD->mark_bar <= gTop));
-
-      if ( unlikely(LD->alerted) )
-      {
-#ifdef O_DEBUGGER
-	if ( debugstatus.debugging && !debugstatus.suspendTrace  )
-	{ LocalFrame fr = dbgRedoFrame(FR, CHP_CLAUSE PASS_LD);
-
-	  if ( fr )
-	  { int action;
-
-	    SAVE_REGISTERS(qid);
-	    clearLocalVariablesFrame(FR);
-	    action = tracePort(fr, BFR, REDO_PORT, NULL PASS_LD);
-	    LOAD_REGISTERS(qid);
-	    ch = BFR;			/* can be shifted */
-
-	    switch( action )
-	    { case ACTION_FAIL:
-		FRAME_FAILED;
-	      case ACTION_IGNORE:
-		VMI_GOTO(I_EXIT);
-	      case ACTION_RETRY:
-		goto retry_continue;
-	      case ACTION_ABORT:
-		THROW_EXCEPTION;
-	    }
-	  }
-	}
-#endif
-#ifdef O_INFERENCE_LIMIT
-        if ( LD->statistics.inferences >= LD->inference_limit.limit )
-	{ SAVE_REGISTERS(qid);
-	  raiseInferenceLimitException();
-	  LOAD_REGISTERS(qid);
-	  if ( exception_term )
-	    THROW_EXCEPTION;
-	}
-#endif
-        Profile(profRedo(ch->prof_node PASS_LD));
-      }
-
-      if ( chp.cref )
-      { ch = newChoice(CHP_CLAUSE, FR PASS_LD);
-	ch->value.clause = chp;
-      } else if ( unlikely(debugstatus.debugging) )
-      { newChoice(CHP_DEBUG, FR PASS_LD);
-      }
-
-      if ( is_signalled(PASS_LD1) )
-      { SAVE_REGISTERS(qid);
-	handleSignals(PASS_LD1);
-	LOAD_REGISTERS(qid);
-	if ( exception_term )
-	  THROW_EXCEPTION;
-      }
-
-			/* require space for the args of the next frame */
-      ENSURE_LOCAL_SPACE(LOCAL_MARGIN, THROW_EXCEPTION);
-      NEXT_INSTRUCTION;
-    }
-    case CHP_TOP:			/* Query toplevel */
-    { DEBUG(MSG_BACKTRACK,
-	    Sdprintf("    REDO #%ld: %s: TOP\n",
-		     loffset(FR),
-		     predicateName(DEF)));
-      DiscardMark(ch->mark);
-      Profile(profRedo(ch->prof_node PASS_LD));
-      QF = QueryFromQid(qid);
-      set(QF, PL_Q_DETERMINISTIC);
-      QF->foreign_frame = PL_open_foreign_frame();
-      assert(LD->exception.throw_environment == &throw_env);
-      LD->exception.throw_environment = throw_env.parent;
-      fail;
-    }
-    case CHP_CATCH:			/* catch/3 & setup_call_cleanup/3 */
-      DEBUG(MSG_BACKTRACK,
-	    Sdprintf("    REDO #%ld: %s: CATCH\n",
-		     loffset(FR),
-		     predicateName(DEF)));
-            if ( true(ch->frame, FR_WATCHED) )
-      { DiscardMark(ch->mark);
-	environment_frame = FR = ch->frame;
-	lTop = (LocalFrame)(ch+1);
-	FR->clause = NULL;
-	if ( true(ch->frame, FR_CLEANUP) )
-	{ SAVE_REGISTERS(qid);
-	  callCleanupHandler(ch->frame, FINISH_FAIL PASS_LD);
-	  LOAD_REGISTERS(qid);
-	} else
-	{ set(ch->frame, FR_CATCHED);
-	}
-	ch = BFR;			/* can be shifted */
-	if ( exception_term )
-	  THROW_EXCEPTION;
-      } else
-      { set(ch->frame, FR_CATCHED);
-      }
-      /*FALLTHROUGH*/
-    case CHP_DEBUG:			/* Just for debugging purposes */
-      DEBUG(MSG_BACKTRACK,
-	    Sdprintf("    REDO #%ld: %s: DEBUG\n",
-		     loffset(FR),
-		     predicateName(DEF)));
-      #ifdef O_DEBUGGER
-      ch0_ref = consTermRef(ch);
-#endif
-      BFR = ch->parent;
-      DiscardMark(ch->mark);
-      goto next_choice;
-  }
-}
   assert(0);
   return FALSE;
 } /* end of PL_next_solution() */
